@@ -1,4 +1,3 @@
-
 #include <string.h>
 #include <limits.h>
 #include <math.h>
@@ -16,7 +15,15 @@ struct crush_map *crush_create()
 {
 	struct crush_map *m;
 	m = malloc(sizeof(*m));
+        if (!m)
+                return NULL;
 	memset(m, 0, sizeof(*m));
+
+	/* initialize legacy tunable values */
+	m->choose_local_tries = 2;
+	m->choose_local_fallback_tries = 5;
+	m->choose_total_tries = 19;
+	m->chooseleaf_descend_once = 0;
 	return m;
 }
 
@@ -48,7 +55,6 @@ void crush_finalize(struct crush_map *map)
 int crush_add_rule(struct crush_map *map, struct crush_rule *rule, int ruleno)
 {
 	__u32 r;
-	int oldsize;
 
 	if (ruleno < 0) {
 		for (r=0; r < map->max_rules; r++)
@@ -61,9 +67,15 @@ int crush_add_rule(struct crush_map *map, struct crush_rule *rule, int ruleno)
 
 	if (r >= map->max_rules) {
 		/* expand array */
+		int oldsize;
+		void *_realloc = NULL;
 		oldsize = map->max_rules;
 		map->max_rules = r+1;
-		map->rules = realloc(map->rules, map->max_rules * sizeof(map->rules[0]));
+		if ((_realloc = realloc(map->rules, map->max_rules * sizeof(map->rules[0]))) == NULL) {
+			return -ENOMEM; 
+		} else {
+			map->rules = _realloc;
+		} 
 		memset(map->rules + oldsize, 0, (map->max_rules-oldsize) * sizeof(map->rules[0]));
 	}
 
@@ -76,6 +88,8 @@ struct crush_rule *crush_make_rule(int len, int ruleset, int type, int minsize,
 {
 	struct crush_rule *rule;
 	rule = malloc(crush_rule_size(len));
+        if (!rule)
+                return NULL;
 	rule->len = len;
 	rule->mask.ruleset = ruleset;
 	rule->mask.type = type;
@@ -126,13 +140,18 @@ int crush_add_bucket(struct crush_map *map,
 			map->max_buckets *= 2;
 		else
 			map->max_buckets = 8;
-		map->buckets = realloc(map->buckets, map->max_buckets * sizeof(map->buckets[0]));
+		void *_realloc = NULL;
+		if ((_realloc = realloc(map->buckets, map->max_buckets * sizeof(map->buckets[0]))) == NULL) {
+			return -ENOMEM; 
+		} else {
+			map->buckets = _realloc;
+		}
 		memset(map->buckets + oldsize, 0, (map->max_buckets-oldsize) * sizeof(map->buckets[0]));
 	}
 
 	assert(map->buckets[pos] == 0);
 
-   	/* add it */
+        /* add it */
 	bucket->id = id;
 	map->buckets[pos] = bucket;
 
@@ -160,21 +179,37 @@ crush_make_uniform_bucket(int hash, int type, int size,
 	struct crush_bucket_uniform *bucket;
 
 	bucket = malloc(sizeof(*bucket));
+        if (!bucket)
+                return NULL;
 	memset(bucket, 0, sizeof(*bucket));
 	bucket->h.alg = CRUSH_BUCKET_UNIFORM;
 	bucket->h.hash = hash;
 	bucket->h.type = type;
 	bucket->h.size = size;
+
+	if (crush_multiplication_is_unsafe(size, item_weight))
+                goto err;
+
 	bucket->h.weight = size * item_weight;
-
 	bucket->item_weight = item_weight;
+	bucket->h.items = malloc(sizeof(__s32)*size);
 
-	bucket->h.items = malloc(sizeof(__u32)*size);
-	bucket->h.perm = malloc(sizeof(__u32)*size);
+        if (!bucket->h.items)
+                goto err;
+
+        bucket->h.perm = malloc(sizeof(__u32)*size);
+
+        if (!bucket->h.perm)
+                goto err;
 	for (i=0; i<size; i++)
 		bucket->h.items[i] = items[i];
 
 	return bucket;
+err:
+        free(bucket->h.perm);
+        free(bucket->h.items);
+        free(bucket);
+        return NULL;
 }
 
 
@@ -190,20 +225,36 @@ crush_make_list_bucket(int hash, int type, int size,
 	struct crush_bucket_list *bucket;
 
 	bucket = malloc(sizeof(*bucket));
+        if (!bucket)
+                return NULL;
 	memset(bucket, 0, sizeof(*bucket));
 	bucket->h.alg = CRUSH_BUCKET_LIST;
 	bucket->h.hash = hash;
 	bucket->h.type = type;
 	bucket->h.size = size;
 
-	bucket->h.items = malloc(sizeof(__u32)*size);
+	bucket->h.items = malloc(sizeof(__s32)*size);
+        if (!bucket->h.items)
+                goto err;
 	bucket->h.perm = malloc(sizeof(__u32)*size);
-	bucket->item_weights = malloc(sizeof(__u32)*size);
+        if (!bucket->h.perm)
+                goto err;
+
+
+        bucket->item_weights = malloc(sizeof(__u32)*size);
+        if (!bucket->item_weights)
+                goto err;
 	bucket->sum_weights = malloc(sizeof(__u32)*size);
+        if (!bucket->sum_weights)
+                goto err;
 	w = 0;
 	for (i=0; i<size; i++) {
 		bucket->h.items[i] = items[i];
 		bucket->item_weights[i] = weights[i];
+
+		if (crush_addition_is_unsafe(w, weights[i]))
+                        goto err;
+
 		w += weights[i];
 		bucket->sum_weights[i] = w;
 		/*printf("pos %d item %d weight %d sum %d\n",
@@ -213,6 +264,13 @@ crush_make_list_bucket(int hash, int type, int size,
 	bucket->h.weight = w;
 
 	return bucket;
+err:
+        free(bucket->sum_weights);
+        free(bucket->item_weights);
+        free(bucket->h.perm);
+        free(bucket->h.items);
+        free(bucket);
+        return NULL;
 }
 
 
@@ -260,22 +318,31 @@ crush_make_tree_bucket(int hash, int type, int size,
 	int i, j;
 
 	bucket = malloc(sizeof(*bucket));
+        if (!bucket)
+                return NULL;
 	memset(bucket, 0, sizeof(*bucket));
 	bucket->h.alg = CRUSH_BUCKET_TREE;
 	bucket->h.hash = hash;
 	bucket->h.type = type;
 	bucket->h.size = size;
 
-	bucket->h.items = malloc(sizeof(__u32)*size);
+	bucket->h.items = malloc(sizeof(__s32)*size);
+        if (!bucket->h.items)
+                goto err;
 	bucket->h.perm = malloc(sizeof(__u32)*size);
+        if (!bucket->h.perm)
+                goto err;
 
 	/* calc tree depth */
 	depth = calc_depth(size);
 	bucket->num_nodes = 1 << depth;
 	printf("size %d depth %d nodes %d\n", size, depth, bucket->num_nodes);
-	bucket->node_weights = malloc(sizeof(__u32)*bucket->num_nodes);
 
-	memset(bucket->h.items, 0, sizeof(__u32)*bucket->h.size);
+        bucket->node_weights = malloc(sizeof(__u32)*bucket->num_nodes);
+        if (!bucket->node_weights)
+                goto err;
+
+	memset(bucket->h.items, 0, sizeof(__s32)*bucket->h.size);
 	memset(bucket->node_weights, 0, sizeof(__u32)*bucket->num_nodes);
 
 	for (i=0; i<size; i++) {
@@ -283,9 +350,17 @@ crush_make_tree_bucket(int hash, int type, int size,
 		node = crush_calc_tree_node(i);
 		printf("item %d node %d weight %d\n", i, node, weights[i]);
 		bucket->node_weights[node] = weights[i];
+
+		if (crush_addition_is_unsafe(bucket->h.weight, weights[i]))
+                        goto err;
+
 		bucket->h.weight += weights[i];
 		for (j=1; j<depth; j++) {
 			node = parent(node);
+
+                        if (crush_addition_is_unsafe(bucket->node_weights[node], weights[i]))
+                                goto err;
+
 			bucket->node_weights[node] += weights[i];
 			printf(" node %d weight %d\n", node, bucket->node_weights[node]);
 		}
@@ -293,6 +368,12 @@ crush_make_tree_bucket(int hash, int type, int size,
 	BUG_ON(bucket->node_weights[bucket->num_nodes/2] != bucket->h.weight);
 
 	return bucket;
+err:
+        free(bucket->node_weights);
+        free(bucket->h.perm);
+        free(bucket->h.items);
+        free(bucket);
+        return NULL;
 }
 
 
@@ -310,6 +391,8 @@ int crush_calc_straw(struct crush_bucket_straw *bucket)
 
 	/* reverse sort by weight (simple insertion sort) */
 	reverse = malloc(sizeof(int) * size);
+        if (!reverse)
+                return -ENOMEM;
 	if (size)
 		reverse[0] = 0;
 	for (i=1; i<size; i++) {
@@ -385,27 +468,45 @@ crush_make_straw_bucket(int hash,
 	int i;
 
 	bucket = malloc(sizeof(*bucket));
+        if (!bucket)
+                return NULL;
 	memset(bucket, 0, sizeof(*bucket));
 	bucket->h.alg = CRUSH_BUCKET_STRAW;
 	bucket->h.hash = hash;
 	bucket->h.type = type;
 	bucket->h.size = size;
 
-	bucket->h.items = malloc(sizeof(__u32)*size);
+        bucket->h.items = malloc(sizeof(__s32)*size);
+        if (!bucket->h.items)
+                goto err;
 	bucket->h.perm = malloc(sizeof(__u32)*size);
+        if (!bucket->h.perm)
+                goto err;
 	bucket->item_weights = malloc(sizeof(__u32)*size);
-	bucket->straws = malloc(sizeof(__u32)*size);
+        if (!bucket->item_weights)
+                goto err;
+        bucket->straws = malloc(sizeof(__u32)*size);
+        if (!bucket->straws)
+                goto err;
 
-	bucket->h.weight = 0;
+        bucket->h.weight = 0;
 	for (i=0; i<size; i++) {
 		bucket->h.items[i] = items[i];
 		bucket->h.weight += weights[i];
 		bucket->item_weights[i] = weights[i];
 	}
 
-	crush_calc_straw(bucket);
+        if (crush_calc_straw(bucket) < 0)
+                goto err;
 
 	return bucket;
+err:
+        free(bucket->straws);
+        free(bucket->item_weights);
+        free(bucket->h.perm);
+        free(bucket->h.items);
+        free(bucket);
+        return NULL;
 }
 
 
@@ -442,33 +543,70 @@ crush_make_bucket(int alg, int hash, int type, int size,
 
 int crush_add_uniform_bucket_item(struct crush_bucket_uniform *bucket, int item, int weight)
 {
-	int newsize = bucket->h.size + 1;
-
-	bucket->h.items = realloc(bucket->h.items, sizeof(__u32)*newsize);
-	bucket->h.perm = realloc(bucket->h.perm, sizeof(__u32)*newsize);
+        int newsize = bucket->h.size + 1;
+	void *_realloc = NULL;
+	
+	if ((_realloc = realloc(bucket->h.items, sizeof(__s32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.items = _realloc;
+	}
+	if ((_realloc = realloc(bucket->h.perm, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.perm = _realloc;
+	}
 
 	bucket->h.items[newsize-1] = item;
-	bucket->h.weight += weight;
-	bucket->h.size++;
 
-	return 0;
+        if (crush_addition_is_unsafe(bucket->h.weight, weight))
+                return -ERANGE;
+
+        bucket->h.weight += weight;
+        bucket->h.size++;
+
+        return 0;
 }
 
 int crush_add_list_bucket_item(struct crush_bucket_list *bucket, int item, int weight)
 {
-	int newsize = bucket->h.size + 1;
-
-	bucket->h.items = realloc(bucket->h.items, sizeof(__u32)*newsize);
-	bucket->h.perm = realloc(bucket->h.perm, sizeof(__u32)*newsize);
-	bucket->item_weights = realloc(bucket->item_weights, sizeof(__u32)*newsize);
-	bucket->sum_weights = realloc(bucket->sum_weights, sizeof(__u32)*newsize);
+        int newsize = bucket->h.size + 1;
+	void *_realloc = NULL;
 
+	if ((_realloc = realloc(bucket->h.items, sizeof(__s32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.items = _realloc;
+	}
+	if ((_realloc = realloc(bucket->h.perm, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.perm = _realloc;
+	}
+	if ((_realloc = realloc(bucket->item_weights, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->item_weights = _realloc;
+	}
+	if ((_realloc = realloc(bucket->sum_weights, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->sum_weights = _realloc;
+	}
+	
 	bucket->h.items[newsize-1] = item;
 	bucket->item_weights[newsize-1] = weight;
-	if (newsize > 1)
-		bucket->sum_weights[newsize-1] = bucket->sum_weights[newsize-2] + weight;
-	else
-		bucket->sum_weights[newsize-1] = weight;
+	if (newsize > 1) {
+
+                if (crush_addition_is_unsafe(bucket->sum_weights[newsize-2], weight))
+                        return -ERANGE;
+
+                bucket->sum_weights[newsize-1] = bucket->sum_weights[newsize-2] + weight;
+	}
+
+        else {
+                bucket->sum_weights[newsize-1] = weight;
+        }
 
 	bucket->h.weight += weight;
 	bucket->h.size++;
@@ -481,23 +619,46 @@ int crush_add_tree_bucket_item(struct crush_bucket_tree *bucket, int item, int w
 	int depth = calc_depth(newsize);;
 	int node;
 	int j;
+	void *_realloc = NULL;
 
 	bucket->num_nodes = 1 << depth;
-	bucket->h.items = realloc(bucket->h.items, sizeof(__u32)*newsize);
-	bucket->h.perm = realloc(bucket->h.perm, sizeof(__u32)*newsize);
-	bucket->node_weights = realloc(bucket->node_weights, sizeof(__u32)*bucket->num_nodes);
+
+	if ((_realloc = realloc(bucket->h.items, sizeof(__s32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.items = _realloc;
+	}
+	if ((_realloc = realloc(bucket->h.perm, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.perm = _realloc;
+	}
+	if ((_realloc = realloc(bucket->node_weights, sizeof(__u32)*bucket->num_nodes)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->node_weights = _realloc;
+	}
 	
 	node = crush_calc_tree_node(newsize-1);
 	bucket->node_weights[node] = weight;
 
 	for (j=1; j<depth; j++) {
 		node = parent(node);
+
+                if (!crush_addition_is_unsafe(bucket->node_weights[node], weight))
+                        return -ERANGE;
+
 		bucket->node_weights[node] += weight;
-		printf(" node %d weight %d\n", node, bucket->node_weights[node]);
+                printf(" node %d weight %d\n", node, bucket->node_weights[node]);
 	}
+
+
+	if (crush_addition_is_unsafe(bucket->h.weight, weight))
+                return -ERANGE;
 	
-	bucket->h.weight += weight;
-	bucket->h.size++;
+        bucket->h.weight += weight;
+        bucket->h.size++;
+
 	return 0;
 }
 
@@ -505,16 +666,37 @@ int crush_add_straw_bucket_item(struct crush_bucket_straw *bucket, int item, int
 {
 	int newsize = bucket->h.size + 1;
 	
-	bucket->h.items = realloc(bucket->h.items, sizeof(__u32)*newsize);
-	bucket->h.perm = realloc(bucket->h.perm, sizeof(__u32)*newsize);
-	bucket->item_weights = realloc(bucket->item_weights, sizeof(__u32)*newsize);
-	bucket->straws = realloc(bucket->straws, sizeof(__u32)*newsize);
+	void *_realloc = NULL;
+
+	if ((_realloc = realloc(bucket->h.items, sizeof(__s32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.items = _realloc;
+	}
+	if ((_realloc = realloc(bucket->h.perm, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.perm = _realloc;
+	}
+	if ((_realloc = realloc(bucket->item_weights, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->item_weights = _realloc;
+	}
+	if ((_realloc = realloc(bucket->straws, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->straws = _realloc;
+	}
 
 	bucket->h.items[newsize-1] = item;
 	bucket->item_weights[newsize-1] = weight;
 
-	bucket->h.weight += weight;
-	bucket->h.size++;
+	if (crush_addition_is_unsafe(bucket->h.weight, weight))
+                return -ERANGE;
+
+	 bucket->h.weight += weight;
+	 bucket->h.size++;
 	
 	return crush_calc_straw(bucket);
 }
@@ -545,6 +727,7 @@ int crush_remove_uniform_bucket_item(struct crush_bucket_uniform *bucket, int it
 {
 	unsigned i, j;
 	int newsize;
+	void *_realloc = NULL;
 	
 	for (i = 0; i < bucket->h.size; i++)
 		if (bucket->h.items[i] == item)
@@ -557,8 +740,16 @@ int crush_remove_uniform_bucket_item(struct crush_bucket_uniform *bucket, int it
 	newsize = --bucket->h.size;
 	bucket->h.weight -= bucket->item_weight;
 
-	bucket->h.items = realloc(bucket->h.items, sizeof(__u32)*newsize);
-	bucket->h.perm = realloc(bucket->h.perm, sizeof(__u32)*newsize);
+	if ((_realloc = realloc(bucket->h.items, sizeof(__s32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.items = _realloc;
+	}
+	if ((_realloc = realloc(bucket->h.perm, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.perm = _realloc;
+	}
 	return 0;
 }
 
@@ -583,10 +774,28 @@ int crush_remove_list_bucket_item(struct crush_bucket_list *bucket, int item)
 	bucket->h.weight -= weight;
 	newsize = --bucket->h.size;
 	
-	bucket->h.items = realloc(bucket->h.items, sizeof(__u32)*newsize);
-	bucket->h.perm = realloc(bucket->h.perm, sizeof(__u32)*newsize);
-	bucket->item_weights = realloc(bucket->item_weights, sizeof(__u32)*newsize);
-	bucket->sum_weights = realloc(bucket->sum_weights, sizeof(__u32)*newsize);
+	void *_realloc = NULL;
+
+	if ((_realloc = realloc(bucket->h.items, sizeof(__s32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.items = _realloc;
+	}
+	if ((_realloc = realloc(bucket->h.perm, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.perm = _realloc;
+	}
+	if ((_realloc = realloc(bucket->item_weights, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->item_weights = _realloc;
+	}
+	if ((_realloc = realloc(bucket->sum_weights, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->sum_weights = _realloc;
+	}
 	return 0;
 }
 
@@ -630,15 +839,29 @@ int crush_remove_tree_bucket_item(struct crush_bucket_tree *bucket, int item)
 	if (newsize != bucket->h.size) {
 		int olddepth, newdepth;
 
-		bucket->h.items = realloc(bucket->h.items, sizeof(__u32)*newsize);
-		bucket->h.perm = realloc(bucket->h.perm, sizeof(__u32)*newsize);
+		void *_realloc = NULL;
+
+		if ((_realloc = realloc(bucket->h.items, sizeof(__s32)*newsize)) == NULL) {
+			return -ENOMEM;
+		} else {
+			bucket->h.items = _realloc;
+		}
+		if ((_realloc = realloc(bucket->h.perm, sizeof(__u32)*newsize)) == NULL) {
+			return -ENOMEM;
+		} else {
+			bucket->h.perm = _realloc;
+		}
 
 		olddepth = calc_depth(bucket->h.size);
 		newdepth = calc_depth(newsize);
 		if (olddepth != newdepth) {
 			bucket->num_nodes = 1 << newdepth;
-			bucket->node_weights = realloc(bucket->node_weights,
-						       sizeof(__u32)*bucket->num_nodes);
+			if ((_realloc = realloc(bucket->node_weights, 
+						sizeof(__u32)*bucket->num_nodes)) == NULL) {
+				return -ENOMEM;
+			} else {
+				bucket->node_weights = _realloc;
+			}
 		}
 
 		bucket->h.size = newsize;
@@ -665,10 +888,28 @@ int crush_remove_straw_bucket_item(struct crush_bucket_straw *bucket, int item)
 	if (i == bucket->h.size)
 		return -ENOENT;
 	
-	bucket->h.items = realloc(bucket->h.items, sizeof(__u32)*newsize);
-	bucket->h.perm = realloc(bucket->h.perm, sizeof(__u32)*newsize);
-	bucket->item_weights = realloc(bucket->item_weights, sizeof(__u32)*newsize);
-	bucket->straws = realloc(bucket->straws, sizeof(__u32)*newsize);
+	void *_realloc = NULL;
+
+	if ((_realloc = realloc(bucket->h.items, sizeof(__s32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.items = _realloc;
+	}
+	if ((_realloc = realloc(bucket->h.perm, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->h.perm = _realloc;
+	}
+	if ((_realloc = realloc(bucket->item_weights, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->item_weights = _realloc;
+	}
+	if ((_realloc = realloc(bucket->straws, sizeof(__u32)*newsize)) == NULL) {
+		return -ENOMEM;
+	} else {
+		bucket->straws = _realloc;
+	}
 
 	return crush_calc_straw(bucket);
 }
@@ -759,6 +1000,7 @@ int crush_adjust_straw_bucket_item_weight(struct crush_bucket_straw *bucket, int
 {
 	unsigned idx;
 	int diff;
+        int r;
 
 	for (idx = 0; idx < bucket->h.size; idx++)
 		if (bucket->h.items[idx] == item)
@@ -770,7 +1012,9 @@ int crush_adjust_straw_bucket_item_weight(struct crush_bucket_straw *bucket, int
 	bucket->item_weights[idx] = weight;
 	bucket->h.weight += diff;
 
-	crush_calc_straw(bucket);
+	r = crush_calc_straw(bucket);
+        if (r < 0)
+                return r;
 
 	return diff;
 }
@@ -798,7 +1042,7 @@ int crush_bucket_adjust_item_weight(struct crush_bucket *b, int item, int weight
 
 /************************************************/
 
-int crush_reweight_uniform_bucket(struct crush_map *crush, struct crush_bucket_uniform *bucket)
+static int crush_reweight_uniform_bucket(struct crush_map *crush, struct crush_bucket_uniform *bucket)
 {
 	unsigned i;
 	unsigned sum = 0, n = 0, leaves = 0;
@@ -808,6 +1052,10 @@ int crush_reweight_uniform_bucket(struct crush_map *crush, struct crush_bucket_u
 		if (id < 0) {
 			struct crush_bucket *c = crush->buckets[-1-id];
 			crush_reweight_bucket(crush, c);
+
+			if (crush_addition_is_unsafe(sum, c->weight))
+                                return -ERANGE;
+
 			sum += c->weight;
 			n++;
 		} else {
@@ -822,7 +1070,7 @@ int crush_reweight_uniform_bucket(struct crush_map *crush, struct crush_bucket_u
 	return 0;
 }
 
-int crush_reweight_list_bucket(struct crush_map *crush, struct crush_bucket_list *bucket)
+static int crush_reweight_list_bucket(struct crush_map *crush, struct crush_bucket_list *bucket)
 {
 	unsigned i;
 
@@ -834,12 +1082,17 @@ int crush_reweight_list_bucket(struct crush_map *crush, struct crush_bucket_list
 			crush_reweight_bucket(crush, c);
 			bucket->item_weights[i] = c->weight;
 		}
+
+		if (crush_addition_is_unsafe(bucket->h.weight, bucket->item_weights[i]))
+                        return -ERANGE;
+
 		bucket->h.weight += bucket->item_weights[i];
 	}
+
 	return 0;
 }
 
-int crush_reweight_tree_bucket(struct crush_map *crush, struct crush_bucket_tree *bucket)
+static int crush_reweight_tree_bucket(struct crush_map *crush, struct crush_bucket_tree *bucket)
 {
 	unsigned i;
 
@@ -852,12 +1105,19 @@ int crush_reweight_tree_bucket(struct crush_map *crush, struct crush_bucket_tree
 			crush_reweight_bucket(crush, c);
 			bucket->node_weights[node] = c->weight;
 		}
+
+		if (crush_addition_is_unsafe(bucket->h.weight, bucket->node_weights[node]))
+                        return -ERANGE;
+
 		bucket->h.weight += bucket->node_weights[node];
+
+
 	}
+
 	return 0;
 }
 
-int crush_reweight_straw_bucket(struct crush_map *crush, struct crush_bucket_straw *bucket)
+static int crush_reweight_straw_bucket(struct crush_map *crush, struct crush_bucket_straw *bucket)
 {
 	unsigned i;
 
@@ -869,8 +1129,13 @@ int crush_reweight_straw_bucket(struct crush_map *crush, struct crush_bucket_str
 			crush_reweight_bucket(crush, c);
 			bucket->item_weights[i] = c->weight;
 		}
-		bucket->h.weight += bucket->item_weights[i];
+
+                if (crush_addition_is_unsafe(bucket->h.weight, bucket->item_weights[i]))
+                        return -ERANGE;
+
+                bucket->h.weight += bucket->item_weights[i];
 	}
+
 	return 0;
 }
 
