@@ -34,6 +34,7 @@ namespace librbd {
       snap_id(CEPH_NOSNAP),
       snap_exists(true),
       read_only(ro),
+      flush_encountered(false),
       exclusive_locked(false),
       name(image_name),
       wctx(NULL),
@@ -68,15 +69,28 @@ namespace librbd {
 
     if (cct->_conf->rbd_cache) {
       Mutex::Locker l(cache_lock);
-      ldout(cct, 20) << "enabling writeback caching..." << dendl;
+      ldout(cct, 20) << "enabling caching..." << dendl;
       writeback_handler = new LibrbdWriteback(this, cache_lock);
+
+      uint64_t init_max_dirty = cct->_conf->rbd_cache_max_dirty;
+      if (cct->_conf->rbd_cache_writethrough_until_flush)
+	init_max_dirty = 0;
+      ldout(cct, 20) << "Initial cache settings:"
+		     << " size=" << cct->_conf->rbd_cache_size
+		     << " num_objects=" << 10
+		     << " max_dirty=" << init_max_dirty
+		     << " target_dirty=" << cct->_conf->rbd_cache_target_dirty
+		     << " max_dirty_age="
+		     << cct->_conf->rbd_cache_max_dirty_age << dendl;
+
       object_cacher = new ObjectCacher(cct, pname, *writeback_handler, cache_lock,
 				       NULL, NULL,
 				       cct->_conf->rbd_cache_size,
 				       10,  /* reset this in init */
-				       cct->_conf->rbd_cache_max_dirty,
+				       init_max_dirty,
 				       cct->_conf->rbd_cache_target_dirty,
-				       cct->_conf->rbd_cache_max_dirty_age);
+				       cct->_conf->rbd_cache_max_dirty_age,
+				       cct->_conf->rbd_cache_block_writes_upfront);
       object_set = new ObjectCacher::ObjectSet(NULL, data_ctx.get_id(), 0);
       object_set->return_enoent = true;
       object_cacher->start();
@@ -206,6 +220,8 @@ namespace librbd {
     plb.add_u64_counter(l_librbd_aio_discard, "aio_discard");
     plb.add_u64_counter(l_librbd_aio_discard_bytes, "aio_discard_bytes");
     plb.add_time_avg(l_librbd_aio_discard_latency, "aio_discard_latency");
+    plb.add_u64_counter(l_librbd_aio_flush, "aio_flush");
+    plb.add_time_avg(l_librbd_aio_flush_latency, "aio_flush_latency");
     plb.add_u64_counter(l_librbd_snap_create, "snap_create");
     plb.add_u64_counter(l_librbd_snap_remove, "snap_remove");
     plb.add_u64_counter(l_librbd_snap_rollback, "snap_rollback");
@@ -255,7 +271,7 @@ namespace librbd {
   {
     map<string, SnapInfo>::const_iterator it;
 
-    for (it = snaps_by_name.begin(); it != snaps_by_name.end(); it++) {
+    for (it = snaps_by_name.begin(); it != snaps_by_name.end(); ++it) {
       if (it->second.id == in_snap_id) {
 	*out_snap_name = it->first;
 	return 0;
@@ -268,7 +284,7 @@ namespace librbd {
   {
     map<string, SnapInfo>::iterator it;
 
-    for (it = snaps_by_name.begin(); it != snaps_by_name.end(); it++) {
+    for (it = snaps_by_name.begin(); it != snaps_by_name.end(); ++it) {
       if (it->second.id == in_snap_id) {
 	*out_pspec = it->second.parent.spec;
 	return 0;
@@ -459,7 +475,7 @@ namespace librbd {
   }
 
   void ImageCtx::write_to_cache(object_t o, bufferlist& bl, size_t len,
-				uint64_t off) {
+				uint64_t off, Context *onfinish) {
     snap_lock.get_read();
     ObjectCacher::OSDWrite *wr = object_cacher->prepare_write(snapc, bl,
 							      utime_t(), 0);
@@ -470,7 +486,7 @@ namespace librbd {
     wr->extents.push_back(extent);
     {
       Mutex::Locker l(cache_lock);
-      object_cacher->writex(wr, object_set, cache_lock);
+      object_cacher->writex(wr, object_set, cache_lock, onfinish);
     }
   }
 
@@ -489,24 +505,45 @@ namespace librbd {
     return r;
   }
 
+  void ImageCtx::user_flushed() {
+    if (object_cacher && cct->_conf->rbd_cache_writethrough_until_flush) {
+      md_lock.get_read();
+      bool flushed_before = flush_encountered;
+      md_lock.put_read();
+
+      uint64_t max_dirty = cct->_conf->rbd_cache_max_dirty;
+      if (!flushed_before && max_dirty > 0) {
+	md_lock.get_write();
+	flush_encountered = true;
+	md_lock.put_write();
+
+	ldout(cct, 10) << "saw first user flush, enabling writeback" << dendl;
+	Mutex::Locker l(cache_lock);
+	object_cacher->set_max_dirty(max_dirty);
+      }
+    }
+  }
+
+  void ImageCtx::flush_cache_aio(Context *onfinish) {
+    cache_lock.Lock();
+    object_cacher->flush_set(object_set, onfinish);
+    cache_lock.Unlock();
+  }
+
   int ImageCtx::flush_cache() {
     int r = 0;
     Mutex mylock("librbd::ImageCtx::flush_cache");
     Cond cond;
     bool done;
     Context *onfinish = new C_SafeCond(&mylock, &cond, &done, &r);
-    cache_lock.Lock();
-    bool already_flushed = object_cacher->flush_set(object_set, onfinish);
-    cache_lock.Unlock();
-    if (!already_flushed) {
-      mylock.Lock();
-      while (!done) {
-	ldout(cct, 20) << "waiting for cache to be flushed" << dendl;
-	cond.Wait(mylock);
-      }
-      mylock.Unlock();
-      ldout(cct, 20) << "finished flushing cache" << dendl;
+    flush_cache_aio(onfinish);
+    mylock.Lock();
+    while (!done) {
+      ldout(cct, 20) << "waiting for cache to be flushed" << dendl;
+      cond.Wait(mylock);
     }
+    mylock.Unlock();
+    ldout(cct, 20) << "finished flushing cache" << dendl;
     return r;
   }
 
